///////////////////////////////////////////////////////////////////////////////////////
// 
// Library for Popular AHRS Algorithms
///////////////////////////////////////////////////////////////////////////////////////

#include "MPU9250.h"
#include "EEPROM.h"
#include "globals.h"
#include "UBLOX.h"
#include "A_ConfigDefines.h"

#include "cTrans.h"
//#include "kfINS6State.h"

#include "Eigen.h"
#include <Eigen/Dense>
void yield() { }

//---- Filter Selection ------
// Set filter type: 
//  0 = Madgwick implementation of Mahoney DCM
//  1 = Madgwick Gradient Descent, in Quaternion form
//  2 = Mahony, in Quaternion form
//  3 = Madwick Original Paper AHRS 
//  4 = DCM Implementation
//  5 = Complimentary Filter
//  6 = uNavEulAHRS

// Select the filter type here
#define MARG 3

// Uncomment this line to disable the magnetometer in the 
// sensor fusion algorithm
//#define DISABLE_MAGN 

#if MARG == 0
  #include "AHRS.h"
#elif MARG == 1
  #include "MadgwickAHRS.h"
#elif MARG == 2
  #include "MahonyAHRS.h"
#elif MARG == 3
  #include "MargUpdateFilter.h"
#elif MARG == 4
  #include "DCM.h"
  DCM dcm;
#elif MARG == 5
  #include "compFlt.h"
#elif MARG == 6
  #include "uNavEulAHRS.h"
  // a uNavAHRS object
  uNavAHRS Filter;
#endif

// an MPU-9250 object on SPI bus 0 with chip select 10
MPU9250 Imu(IMU_BUS, IMU_ADDR);
int status;

UBLOX gps(GPS_Start);
bool newGPSData;
unsigned long prevTOW;
gpsData uBloxData;
//unsigned long tstart, tstop;

void setup() {
  // serial to display data
  cout.begin(cout_BAUD);
  #ifdef coutD
    coutD.begin(coutD_BAUD);
  #endif
  
  // serial to display data
  cout.begin(cout_BAUD);
  #ifdef coutD
    coutD.begin(coutD_BAUD);
  #endif

  while(!cout && millis() < 2000 ) {}
  delay(5000);

  gps.begin(GPS_BAUD);
  if ( 255 != GPS_AltRx ) GPS_PORT.setRX(GPS_AltRx);
  //attachInterrupt(digitalPinToInterrupt(GPS_SRX), GPS_serialrx_isr, RISING);

  //init imu
  // start communication with IMU 
  if ( 255 != IMU_SCL ) Wire.setSCL(IMU_SCL);
  if ( 255 != IMU_SDA ) Wire.setSDA(IMU_SDA);
  status = Imu.begin();
  if ( 0 != IMU_SPD ) Wire.setClock(IMU_SPD);

  // Initalize IMU   
  initIMU();

  // quaternion of sensor frame relative to auxiliary frame
  q[0] = 1.0f; q[1] = 0.0f; 
  q[2] = 0.0f; q[3] = 0.0f;  

  // Time parameter initialization
  _dt = 0.001;
  tsGPS = 0;
  tsIMU = 0;
}

void loop(){

  readGPS(); 
  if(rtk.upDated){
  // Calculate updated utcTime for this loop iteration
  utcTime = float(uBloxData.utcHour)*60.0f*60.0f +
            float(uBloxData.utcMin) *60.0f +
            float(uBloxData.utcSec) +
            float(uBloxData.utcNano)/1000000000.0;
    timeStamp = 0.0f;
    //serialDataOut();  // Print function using printf statements
    //serialPortOut();  // String print function for Telemetry Viewer
  } else {
    //utcTime = 0.0f;
  }
  
  // if we are tracking 5 GPS satellites, begin AHRS and INS filtering
  if(rtk.numSV > 4){

    
  // Wait for new data from IMU to run filter algorithms
  if (newIMUData == 1) {
    newIMUData = 0;
    
    // read the sensor
    // Imu.readSensor();  //BUGBUG DUPLICATE

    // Function cal to get IMU calibrated data +
    // filter results (quaternions)
Dtime( 10, 0, 0 );
    getIMU();
Dtime( 10, 10, "getIMU" );
  readGPS(); 
        
    lastUpdate = now;
    
    if((0==GPS_newData) && (sincePrint > printRate)){
      sincePrint=0;
      
Dtime( 1, 0, 0 );
      #if (MARG == 5)
        //do nothing
      #elif(MARG == 4)
        dcm.getEulerDeg(ypr);
      #elif (MARG == 6)
        ypr[0] = Filter.getYaw_rad()*180.0f/PI;
        ypr[1] = Filter.getPitch_rad()*180.0f/PI;
        ypr[2] = Filter.getRoll_rad()*180.0f/PI;
        //cout.println(Filter.getHeading_rad()*180.0f/PI);
      #else
        getYawPitchRoll(ypr);
      #endif
Dtime( 1, 10, "E°_YPR" );
  readGPS(); 

      // Rotate accel body to accel inertial (NED)
Dtime( 1, 0, 0 );
      body2inertial();
Dtime( 1, 10, "b2i" );

      // KF propagate, update, and reset
  readGPS();
Dtime( 1, 0, 0 );
      kfProp();
Dtime( 1, 10, "kfP__" );
  readGPS(); 
Dtime( 1, 0, 0 );
      kfUpdate();
Dtime( 1, 10, "kf_U_" );
  readGPS(); 
Dtime( 1, 0, 0 );
      kfReset();
Dtime( 1, 10, "kf__R" );
  readGPS(); 

      // Convert ECEF position back to LLA
      xECEF(0,0) = xk1(0,0);
      xECEF(1,0) = xk1(1,0);
      xECEF(2,0) = xk1(2,0);
      
Dtime( 1, 0, 0 );
      RefLLA = ecef2lla(xECEF);
Dtime( 1, 10, "ecef2lla" );

    /*
    Serial.print("KF ECEF: ");
    Serial.print(xk1(0,0));
    Serial.print(",  ");
    Serial.print(xk1(1,0));
    Serial.print(",  ");
    Serial.println(xk1(2,0));
    Serial.print("KK LLA: ");
    Serial.print(RefLLA(0,0)*rad2deg);
    Serial.print(",  ");
    Serial.print(RefLLA(1,0)*rad2deg);
    Serial.print(",  ");
    Serial.println(RefLLA(2,0));
    */

  readGPS();
    // Send data to serial port. Check for duplicate GPS readings first.
    //serialDataOut();  // Print function using printf statements
      serialPortOutKF();  // String print function for Telemetry Viewer

    } // if sincePrint
   } // if NewData for IMU
  } // if uBloxData
} // end main loop


//****************************************************
//
// On Pass through toggle flag and call function, flip 
// on seond in function
//****************************************************
void toggleGPS(){
  if(passThrough_toggle == 0){
   passThrough_toggle = 1;
   passThroughMode();
  } else {
    passThrough_toggle = 0;
  }
}

//****************************************************
//
// Get IMU Sensor and Filter Results
//****************************************************
void getIMU()
{
    // read the sensor
    Imu.readSensor();

    // Initialize Kalman filter
    if (!kfInitFlag){
      delay(500);
      kfInit();
      kfInitFlag = true;
    }
    readGPS();
    // update the filter
    gx1 = Imu.getGyroX_rads(); gy1 = Imu.getGyroY_rads(); gz1 = Imu.getGyroZ_rads();
    ax1 = Imu.getAccelX_mss(); ay1 = Imu.getAccelY_mss(); az1 = Imu.getAccelZ_mss();
    mx1 = Imu.getMagX_uT();    my1 = Imu.getMagY_uT();    mz1 = Imu.getMagZ_uT();  

    //now = micros(); // Set on data ready interrupt 
    #if MARG != 6
      _dt = ((now - lastUpdate) * 0.000001);
    #endif
    readGPS();
     
    #if(MARG == 0 || MARG == 1 || MARG == 2 || MARG == 3)  
      #if (not defined(DISABLE_MAGN))
        #if MARG == 0
            AHRSupdate(gx1, gy1, gz1, ax1, ay1, az1, mx1, my1, mz1);
        #elif MARG == 1
            MadgwickAHRSupdate(gx1, gy1, gz1, ax1, ay1, az1, mx1, my1, mz1);        
        #elif MARG == 2
            MahonyAHRSupdate(gx1, gy1, gz1, ax1, ay1, az1, mx1, my1, mz1);
        #else
            MARGUpdateFilter(gx1, gy1, gz1, ax1, ay1, az1, mx1, my1, mz1);
        #endif
      #else
        #if MARG == 0
            AHRSupdateIMU(gx1, gy1, gz1, ax1, ay1, az1);
        #elif MARG == 1
            MadgwickAHRSupdateIMU(gx1, gy1, gz1, ax1, ay1, az1);
        #else
            MARGUpdateFilterIMU(gx1, gy1, gz1, ax1, ay1, az1);
        #endif
      #endif
    #elif (MARG == 4)
        //val[9] = maghead.iheading(1, 0, 0, val[0], val[1], val[2], val[6], val[7], val[8]);
        magHead = calcMagHeading(q[0], q[1], q[2], q[3], mx1, my1, mz1);
        dcm.setSensorVals(gx1*180./PI, gy1*180./PI, gz1*180./PI, ax1, ay1, az1, mx1, my1, mz1, magHead);
        dcm.G_Dt = _dt;
        dcm.calDCM();
        dcm.getDCM2Q(q);
    #elif (MARG == 5)
        //compFilter(gx1, gy1, gz1, ax1, ay1, az1, 0, 0, 0);    //yaw based on filter alone
        compFilter(gx1, gy1, gz1, ax1, ay1, az1, mx1, my1, mz1);//yaw calculated with magnetometer
    #elif MARG == 6
        upDated = Filter.update(gx1, gy1, gz1, ax1, ay1, az1, mx1, my1, mz1);
    #endif
}

//****************************************************
//****************************************************
//Interupt Control for IMU and GPS

// This routine is called via interrupt from the IMU
void runFilter() {
  now = micros(); // This is when the data reported READY
  newIMUData = 1;
  IMU_RX_time = tsIMU;
}


//========================================
#define RxGapTime ((F_CPU / GPS_BAUD) * 30)
//========================================
// FASTRUN void GPS_serialrx_isr()
// GLOBALS :: volatile uint32_t GPS_RX_Time = 0;
FASTRUN void GPS_serialrx_isr() {
    GPS_RX_time = tsGPS; // Start bit transition detected after 'idle' time
    detachInterrupt(digitalPinToInterrupt(GPS_SRX));
    GPS_newData=1;
}

// Delta time function::
// ID 0 to 9 { 10 {IMU} and 11 {GPS} reserved }
// State :: 
//   0:  zero the ID timer, 1: End show the ID's time, 2: return time no print
//   10: Only show time if it is over 700 uS
// > Times are static - so they persist and unique ID's can be nested
uint32_t Dtime( int ID, int State, const char *szID  ) {
  uint32_t RetVal=micros();
#ifdef coutD
//  uint32_t Tnow = micros(), RetVal;
  static uint32_t TimeVals[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
  if ( 1 == State ) {
    RetVal = RetVal - TimeVals[ ID ];
    coutD.print( szID );
    coutD.print( " uS=");
    coutD.println( RetVal );
  }
  else if ( 10 == State ) {
    RetVal = RetVal - TimeVals[ ID ];
    if ( 750 < RetVal) {
      coutD.print( szID );
      coutD.print( " uS=");
      coutD.println( RetVal );
    }
    else if ( 10 == ID )
      coutD.print( "*");
    else 
      coutD.print( "^");
}
  else if ( 2 == State ) {
    RetVal = RetVal - TimeVals[ ID ];
  }
  else if (0 == State) {
    TimeVals[ ID ] = RetVal;
  }
#endif
  return RetVal;
}

